!
! Module: espfilt_mod
!
! SAS Version 12.0
! Version 1.7 15 March 2012
!
! Authors:  Brendan Perry - NASA/GSFC
!          
! Description: Generates soft proton contamination-filtered products from event lists.
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module espfilt_mod

  use dal
  use param 
  use errorHandling
  use marq_mod
  use f90_unix_proc

  implicit none 
  
  private 

  public espfilt

contains 
  
  !---------------------------------------------------------------------
  subroutine espfilt
  !---------------------------------------------------------------------
!    implicit none

! declare input parameters first
    type(DataSetT) :: objevliset
    type(TableT) :: inevents
    type(ColumnT) :: time_obj
    character(len=1) :: instnum
    character(len=2) :: patt
    character(len=10) :: method, obsid, inst, expid
    character(len=1024) :: evli, root 
    integer :: smoothfactor, binsize, specchanmin, specchanmax, &
        n_evli, i
    logical :: withsmoothing, withbinning, withspecranges  
    real(kind=single) :: ratio
!
! get the parameters from the par file
!
    write(errstr,*) "ESPFILT: Retrieving parameters from par file."
    call message(AppMsg,Noisy)

    n_evli = parameterCount("eventset")
    method = stringParameter("method") ! corner or ratio
    withsmoothing = booleanParameter("withsmoothing")
    smoothfactor = intParameter("smooth")
    ratio = realParameter("ratio") ! flare factor ratio
    withspecranges = booleanParameter("withspecranges")
    specchanmin = intParameter("specchanmin")
    specchanmax = intParameter("specchanmax")
    withbinning = booleanParameter("withbinning")
    binsize = intParameter("binsize")

    if (n_evli .lt. 1) then
       call error ("noEventSet","No initial event file selected")
    end if

    write(errstr,*) "ESPFILT: looping through the ",n_evli," event lists."
    call message(AppMsg,Verbose)
!
    do i=1,n_evli
       evli = stringParameter("eventset",i-1)
       write(errstr,*) "ESPFILT: Processing event list: ",trim(evli)
       call message(AppMsg,Verbose)
       objevliset = dataSet(evli, READ)
       if (.not.hasBlock(objevliset,"EVENTS")) then
          call error("badInfile","Input file does not contain an EVENTS extension.")
       else
          inevents = table(objevliset,"EVENTS")
       endif
       if (.not.hasColumn(inevents,"TIME")) then
          call error("badInfile","Input file incorrect - no TIME column")
       else
          time_obj = column(inevents,"TIME",READ)
       endif

! get some header keywords to create output file's names (root)

       obsid = stringAttribute(objevliset,"OBS_ID")
       inst = stringAttribute(objevliset,"INSTRUME")
       write(errstr,*) "ESPFILT: OBSID=",trim(obsid)," INSTRUME=",inst
       call message(AppMsg,Verbose)
!
       if (inst(:3) .eq. "EMO") then ! EPIC MOS
          instnum = inst(5:5)
          inst = "M"//instnum
          patt = "12"
       elseif (inst(:3) .eq. "EPN") then ! EPIC PN
          inst = "PN"
          instnum = "4"
          patt = "4"
       else
          call error("badInfile",&
               "Cannot process eventlist without INSTRUME EMOSn or EPN")
       end if
       expid = stringAttribute(objevliset,"EXPIDSTR")

! create output file name root
       root = "P"//trim(obsid)//trim(inst)//trim(expid)
       call release(objevliset)
!
! other methods will eventually go here...
!

! For PN observations check the SUBMODE keyword in the event file primary 
! header, if not 'PrimeFullWindow' or 'PrimeFullWindowExtended' then die 
! gracefully and with an explanation

       if (method .eq. "corner") then 
          write(errstr,*) "ESPFILT: Processing using the CORNER method."
          call message(AppMsg,Verbose)
          call corner_method (evli,root,instnum,withbinning,&
               withsmoothing,binsize,smoothfactor,specchanmin,&
               specchanmax,patt)
       elseif (method .eq. "ratio") then 
          write(errstr,*) "ESPFILT: Processing using the RATIO method."
          call message(AppMsg,Verbose)
          call ratio_method (evli,root,instnum,withbinning,&
               withsmoothing,binsize,smoothfactor,specchanmin,&
               specchanmax,ratio,patt)
       else
          call error("UnsupportedMethod",&
               "Only Corner or Ratio methods are supported at this time")
       end if
!
! future
!       call delta()
!       call user_specified()
    end do
    call message(AppMsg,Verbose)
  end subroutine espfilt
!
!--------------------------------------------------
! subroutines
!--------------------------------------------------
!
  subroutine corner_method (evfile,root,instnum,binning,smoothing,&
       binsize,smoothfactor,minchan,maxchan,patt)
!
! corner method works !
!
    character(len=1024),intent(in) :: evfile, root
    character(len=1024) :: objimgF, objlcF, corlcF, gtitxtF
    character(len=1024) :: gtifitsF, objevlifiltF, histqdpF
    character(len=1024) :: objimgfiltF, corimgfiltF, corevlifiltF
    character(len=1024) :: cornerdef, command
    character(len=1) :: instnum
    character(len=2) :: patt
    character(len=5) :: cminchan, cmaxchan
    integer, intent(in) :: minchan, maxchan, smoothfactor,&
         binsize
    integer :: method
    logical, intent(in) :: binning, smoothing
    logical :: f_ok = .false.
    real(kind=double) :: ratio = 1.0
    write(cminchan,'(I5)') minchan
    write(cmaxchan,'(I5)') maxchan

!
! chip "corners" defined for evselect calls later
!
    if (instnum .eq. "1") then ! EMOS1
       cornerdef = "!(CIRCLE(100,-200,17700,DETX,DETY)"//&
            "||CIRCLE(834,135,17100,DETX,DETY)"//&
            "||CIRCLE(770,-803,17100,DETX,DETY)"//&
            "||BOX(-20,-17000,6500,500,0,DETX,DETY)"//&
            "||BOX(5880,-20500,7500,1500,10,DETX,DETY)"//&
            "||BOX(-5920,-20500,7500,1500,350,DETX,DETY)"//&
            "||BOX(-20,-20000,5500,500,0,DETX,DETY)"//&
            "||BOX(-12900,16000,250,4000,0,DETX,DETY)"//&
            "||BOX(80,18600,150,1300,0,DETX,DETY)"//&
            "||BOX(-10,-18800,125,1500,0,DETX,DETY))"
    else if (instnum .eq. "2") then ! EMOS2
       cornerdef = "!(CIRCLE(435,1006,17100,DETX,DETY)"//&
            "||CIRCLE(-34,68,17700,DETX,DETY)"//&
            "||BOX(-20,-17000,6500,500,0,DETX,DETY)"//&
            "||BOX(5880,-20500,7500,1500,10,DETX,DETY)"//&
            "||BOX(-5920,-20500,7500,1500,350,DETX,DETY)"//&
            "||BOX(-20,-20000,5500,500,0,DETX,DETY))"
    else ! EPN
       cornerdef = "!((DETX,DETY) IN circle(-2200,-1090,18100))"
    end if
    write(errstr,*) "CORNER_METHOD: Chip corner definition:",trim(cornerdef)
    call message(AppMsg,Noisy)
!
! evselect runs on OBJECT 
!
    write(errstr,*) "CORNER_METHOD: Running evselect to create "//&
         "*OBJECT IMAGE*"
    call message(AppMsg,Verbose)
    objimgF = trim(root)//"-objimg.FIT"
    command = "evselect table="// trim(evfile) //" withimageset=yes "//&
         " imageset="// trim(objimgF) // " xcolumn=DETX ycolumn=DETY "//&
         "imagebinning=imageSize ximagesize=600 yimagesize=600"
    call system(trim(command))
! 
    write(errstr,*) "CORNER_METHOD: Running evselect to create "//&
         "*OBJECT LIGHTCURVE*"
    call message(AppMsg,Verbose)
    objlcF = trim(root)//"-objlc.FIT"
    cminchan = adjustl(cminchan)
    cmaxchan = adjustl(cmaxchan)
    command = "evselect table="// trim(evfile) //":EVENTS "//&
         "withfilteredset=yes expression='(PATTERN<="// trim(patt) //")&&(PI in "//&
         "["// trim(cminchan) // ":" // trim(cmaxchan) // "])&&"//&
         "((FLAG & 0xfb0000) == 0)' filtertype=expression "//&
         "keepfilteroutput=no withrateset=yes rateset="// trim(objlcF) //&
         " timecolumn='TIME' maketimecolumn=yes makeratecolumn=yes "//&
         "timebinsize=1 withtimeranges=no"
    call system(trim(command))
!
! create corner lightcurve
!
    write(errstr,*) "CORNER_METHOD: Running evselect to create "//&
         "*CORNER LIGHTCURVE*"
    call message(AppMsg,Verbose)
    corlcF = trim(root)//"-corlc.FIT"
! PI min and max not SPECCHANMIN and MAX???? 
    command = "evselect table="// trim(evfile) //":EVENTS "//&
         "withfilteredset=yes expression="//&
         "'(PATTERN<="// trim(patt) //")&&(PI in "//&
         "["// trim(cminchan) // ":" // trim(cmaxchan) // "])&&"//&
         "((FLAG & 0x766a0f63) == 0)&&"// trim(cornerdef) //&
         "' filtertype=expression keepfilteroutput=no "//&
         "withrateset=yes rateset="// trim(corlcF) //&
         " timecolumn='TIME' maketimecolumn=yes makeratecolumn=yes "//&
         "timebinsize=1 withtimeranges=no"
    call system(trim(command))
!
! call PGPLOT directly in the future? Eliminate QDP file construct!!!
!
    histqdpF = trim(root)//"-hist.qdp" ! QDP output file
    write(errstr,*) "CORNER_METHOD: Running Clean_LC..."
    call message(AppMsg,Verbose)
    gtitxtF = trim(root)//"-gti.txt" ! GTI temp text file
    gtifitsF = trim(root)//"-gti.FIT"! GTIBUILD output FITS file
!
! clean_lc is the engine that creates the gti txt file for use by
! gtibuild and the smoothed, binned lc's
!
    method = 1 ! corner = 1, ratio = 2
    call clean_lc(gtitxtF,objlcF,corlcF,objlcF,histqdpF,&
         binning,smoothing,binsize,smoothfactor,method,ratio)
    command = "gtibuild file="// trim(gtitxtF) //" table="// trim(gtifitsF)
    call system(trim(command))
!
! if GTI file exists, create filtered, flare-free event list
!
    inquire(file=trim(gtifitsF),exist=f_ok)
    if (f_ok) then
       objevlifiltF = trim(root)//"-objevlifilt.FIT"
       command = "evselect table="// trim(evfile) //&
            " withfilteredset=yes "//&
            "expression='(PATTERN<="// trim(patt) //")&&GTI("//&
            trim(gtifitsF) //",TIME)&&"//&
            "((FLAG & 0x766a0f63)==0)' filteredset="//&
            trim(objevlifiltF) //&
            " filtertype=expression keepfilteroutput=yes "//&
            "updateexposure=yes filterexposure=yes"
       call system(trim(command))
       objimgfiltF = trim(root)//"-objimgfilt.FIT"
       command = "evselect table="// trim(objevlifiltF) //&
            " withimageset=yes imageset="// trim(objimgfiltF) //&
            " xcolumn=DETX ycolumn=DETY imagebinning=imageSize "//&
            "ximagesize=600 yimagesize=600"
       call system(trim(command))
!
! create corner-only events file 
       corevlifiltF = trim(root) // "-corevlifilt.FIT"
       command = "evselect table="// trim(evfile) //":EVENTS " //&
            "withfilteredset=yes expression='(PATTERN<="// trim(patt) //&
            ")&&GTI("// trim(gtifitsF) //&
            ",TIME)&&((FLAG & 0x766a0f63) == 0)&&"//&
            trim(cornerdef)//"' filteredset="//&
            trim(corevlifiltF)//&
            " filtertype=expression keepfilteroutput=yes "//&
            "updateexposure=yes filterexposure=yes" 
       call system(trim(command))

! create the image of the corner-only events for mos_doc.bat use
       corimgfiltF = trim(root) // "-corimgfilt.FIT"
       command = "evselect table="// trim(corevlifiltF) //&
            " withimageset=yes imageset="// trim(corimgfiltF) //&
            " xcolumn=DETX ycolumn=DETY imagebinning=imageSize "//&
            "ximagesize=600 yimagesize=600"
       call system(trim(command))
       
    else
       call error("noGTI","No GTI FITS file was found, cannot continue")
    end if
    return       
  end subroutine corner_method
!
!
!
  subroutine ratio_method (evfile,root,instnum,binning,smoothing,&
       binsize,smoothfactor,minchan,maxchan,ratio,patt)
!
!  added 05/2008 -- BHP from Kip Kuntz Algorithm SS input.
!
! This algorithm extracts the number of counts per pixel in the 
! user specified (defaulted to 8.0-12.0 keV) band in an annulus 
! in the FOV (13900<R<16900) and compare to the counts per pixel
! in the corner regions for the same energy band.
!
! The ratio of interest is:
! (annulus_counts) * corner_area / (corner_counts * annulus_area)
!
! If this ratio is greater than a user specified value (defaulting
! to 1.2) then there is flaring.
!
! Next, a set of GTIs is created in which flaring is NOT occurring,
! and these are used to created a filtered Event List of the object.
!
    type(DataSetT) :: set
    type(TableT)   :: tab
!    type(ArrayT)   :: arr
    character(len=1024),intent(in) :: evfile, root
    character(len=1024) :: objimgF, objlcF, corlcF, fovannlcF
    character(len=1024) :: objevlifiltF, objimgfiltF,&
         corevlifiltF, corimgfiltF
    character(len=1024) :: fovanndef, cornerdef
    character(len=1024) :: histqdpF, gtitxtF, gtifitsF
    character(len=1024) :: fovannevliF, fovannspecF, corevliF, corspecF
    character(len=1024) :: command    
    character(len=1) :: instnum
    character(len=1) :: patt
    character(len=5) :: cminchan, cmaxchan

    integer, intent(in) :: minchan, maxchan
    integer, intent(in) :: smoothfactor, binsize
    integer :: method
    logical, intent(in) :: binning, smoothing
    logical :: f_ok = .false.
    real(kind=double) :: cor_area, fovann_area, cor_cnt, &
        fovann_cnt, newratio, cor_time, fovann_time, unc
    real(kind=single),intent(in) :: ratio
    write(cminchan,'(I5)') minchan
    write(cmaxchan,'(I5)') maxchan
    cminchan = adjustl(cminchan)
    cmaxchan = adjustl(cmaxchan)
!
! corner_method = 1, ratio_method=2 
!
    method = 2 

    if (binsize .lt. 100) then
       call warning("RatioBinLT100","Binsize for Ratio Method should be >= 100")
    end if

    if (instnum .eq. "1") then                ! EMOS1
!
! same corner def as CORNER_METHOD
!
! 
       write (errstr,*) "RATIO_METHOD: Defining corner for EMOS1"
       call message(AppMsg,Noisy)
       cornerdef = "!(CIRCLE(100,-200,17700,DETX,DETY)" //&
            "||CIRCLE(834,135,17100,DETX,DETY)" //&
            "||CIRCLE(770,-803,17100,DETX,DETY)" //&
            "||BOX(-20,-17000,6500,500,0,DETX,DETY)" //&
            "||BOX(5880,-20500,7500,1500,10,DETX,DETY)" //&
            "||BOX(-5920,-20500,7500,1500,350,DETX,DETY)" //&
            "||BOX(-20,-20000,5500,500,0,DETX,DETY)" //&
            "||BOX(-12900,16000,250,4000,0,DETX,DETY)" //&
            "||BOX(80,18600,150,1300,0,DETX,DETY)" //&
            "||BOX(-10,-18800,125,1500,0,DETX,DETY))"

       write(errstr,*) "RATIO_METHOD: Defining FOV annulus for EPN"
       call message(AppMsg,Noisy)
       fovanndef = "(CIRCLE(100,-200,16900,DETX,DETY)" //&
            "&&!CIRCLE(100,-200,13900,DETX,DETY))" ! the donut

    else if (instnum .eq. "2") then ! EMOS2
       write (errstr,*) "RATIO_METHOD: Defining corner for EMOS2"
       call message(AppMsg,Noisy)
       cornerdef = "!(CIRCLE(435,1006,17100,DETX,DETY)" //&
            "||CIRCLE(-34,68,17700,DETX,DETY)" //&
            "||BOX(-20,-17000,6500,500,0,DETX,DETY)" //&
            "||BOX(5880,-20500,7500,1500,10,DETX,DETY)" //&
            "||BOX(-5920,-20500,7500,1500,350,DETX,DETY)" //&
            "||BOX(-20,-20000,5500,500,0,DETX,DETY))"
       write(errstr,*) "RATIO_METHOD: Defining FOV annulus for EPN"
       call message(AppMsg,Noisy)
       fovanndef = "(CIRCLE(0,-200,16900,DETX,DETY)" //&
            "&&!CIRCLE(0,-200,13900,DETX,DETY))" ! the donut
    else
!
! For the PN fin/fout selection
! Steve's fout = cornerdef = corners of instrument defined
!
        write(errstr,*) "RATIO_METHOD: Defining corner for EPN"
        call message(AppMsg,Noisy)
        cornerdef = "!((DETX,DETY) in circle(-2200,-1100,18080))"
!
! fin = fovanndef == field of view annulus defined
!
        write(errstr,*) "RATIO_METHOD: Defining FOV annulus for EPN"
        call message(AppMsg,Noisy)
        fovanndef = "((DETX,DETY) in circle(-2200,-1100,18080))" //&
        "&&!((DETX,DETY) in circle(-2200,-1100,8200))&&"//&
        "(PATTERN <= "// trim(patt) //")&&((FLAG & 0x762a097c) == 0)" 

! this is the 'donut'

!!! do I use the pattern in all cases???	
!!! Use  from Steve 

    end if

    write(errstr,*) "RATIO_METHOD: Chip corner definition:",trim(cornerdef)
    call message(AppMsg,Sparse)
    write(errstr,*) "RATIO_METHOD: Chip FOV definition:",trim(fovanndef)
    call message(AppMsg,Sparse)
!
!!! put following into subroutine shared by corner and ratio???
!
!
! first create the corner event file (1)
! this is just the corner pixels defined with pattern/flag removal
! full energy range (def 8-12keV
!
!    corevfiltF = trim(root) // "-corevfilt.FIT"  ! kk corn
! 
    corevliF = trim(root) // "-corevli.FIT"       ! kk ehig

    command = "evselect table="// trim(evfile) //":EVENTS" //&
         " withfilteredset=yes expression='(PATTERN<=" // trim(patt) // &
         ")&&((FLAG & 0x766a0f63) == 0)&&" // trim(cornerdef) // &
         "' filteredset="// trim(corevliF) //&
         " filtertype=expression keepfilteroutput=yes"//&
         " updateexposure=yes filterexposure=yes" 

! same as above but with user defined energy range. The EVENT LIST
! can have full energy, as can the Spectrum. The LC will be culled
! on energy. However, spectrum will be selected as well as the other
! events aren't used.
!    command = "evselect table=" // trim(evfile) // ":EVENTS " //&
!         " withfilteredset=yes expression='(PI in [" // trim(cminchan) //&
!         ":" // trim(cmaxchan) // "])&&"// trim(cornerdef) //&
!         "&&(PATTERN<=12)&&((FLAG & 0x766a0f63) == 0)' "//&
!         "' filtertype=expression filteredset=" // trim(corevliF) //&
!         " keepfilteroutput=yes updateexposure=yes filterexposure=yes"

    write(errstr,*)trim(command)
    call message(AppMsg,Noisy)
    call system (trim(command))
!
! kk does the below command, but 8000 and 12000 are now user-selected. 
! he also takes an already-filtered-for-cornerdef eventfile, whereas
! it now takes from a raw event list and does both cornerdef and others.
! command = "evselect table=mos".$ins_num."-corn.fits:EVENTS ".
! "withfilteredset=yes expression='(PI in [8000:12000])'".//&
!  filtertype=expression "   
! command = "evselect table=" // trim(corevfileF) // ":EVENTS " //&
! if use evfile, add cornerdef?
! so, raw eventlist, filter on energy
!
    write(errstr,*) "RATIO_METHOD: Evselect extract corner spectrum"
    call message(AppMsg,Noisy)
    corspecF = trim(root) // "-corspec.FIT"  ! mos_instnum_corn.pi

!    command = "evselect table=mos".$ins_num."-corn.fits:EVENTS "//&
! mos".$ins_num."-corn.fits:EVENTS "//& ??? 
! mos".$ins_num.-corn.pi
! " energycolumn=PI spectralbinsize=15 withspecranges=yes "//&

!
! already applied pattern, flag and cornerdef, so just energy,
! binsize, etc needed:
!
! extract corner spectrum second (2)

    command = "evselect table=" // trim(corevliF) // &
         " withfilteredset=yes expression='TRUE'"//&
!         "'(PATTERN<=12)&&((FLAG & 0x766a0f63) == 0)&&"//&
!         trim(cornerdef) //  
         " filtertype=expression "//&
         "keepfilteroutput=no updateexposure=yes filterexposure=yes "//&
         "withspectrumset=yes spectrumset=" // trim(corspecF) //&
         " energycolumn=PI spectralbinsize=15 withspecranges=yes "//&
         "specchannelmin=0 specchannelmax=11999"
!    command = "evselect table=" // trim(corevliF) // &
!         " withfilteredset=yes expression="//&
!         "'(PATTERN<=12)&&((FLAG & 0x766a0f63) == 0)&&"//&
!         trim(cornerdef) //  "' filtertype=expression "//&
!         "keepfilteroutput=no updateexposure=yes filterexposure=yes "//&
!         "withspectrumset=yes spectrumset=" // trim(corspecF) //&
!         " energycolumn=PI spectralbinsize=15 withspecranges=yes "//&
!         "specchannelmin=0 specchannelmax=11999"
!
! should specchans be set by user? Aren't they redundant? Yes; The
! event list was created with cmin/cmaxchan so no need to even
! have the above...
!
    write(errstr,*)trim(command)
    call message(AppMsg,Noisy)
    call system (trim(command))

! command = "backscale spectrumset=mos".$ins_num.
! "-corn.pi badpixlocation=mos".$ins_num.
! "-clean.fits withbadpixcorr=yes ignoreoutoffov=no";
!
! backscale the corner spectrum created above to get the BACKSCAL
! keyword (the area of the corners used to make the corner spectrum.
! as from the task Backscale: area = geometric_area - ccd_gaps - bad_pixels)
!
! run backscale on the corner spectrum (3)
!
    write(errstr,*) "RATIO_METHOD: Backscale corner spectrum."
    call message(AppMsg,Noisy)
    command = "backscale spectrumset=" // trim(corspecF) //&
         " badpixlocation=" // trim(evfile) //&
         " withbadpixcorr=yes ignoreoutoffov=no";
    write(errstr,*)trim(command)
    call message(AppMsg,Noisy)
    call system (trim(command))
!
! now do 1-3 for the FOV annulus. First the FOV annulus event file (4)
!
    fovannevliF = trim(root) // "-annevli.FIT"       ! kk ehig

    command = "evselect table="// trim(evfile) //":EVENTS" //&
         " withfilteredset=yes expression='(PATTERN<="// trim(patt)// &
         ")&&((FLAG & 0x766a0f63) == 0)&&" // trim(fovanndef)// &
         "' filteredset="// trim(fovannevliF) //&
         " filtertype=expression keepfilteroutput=yes"//&
         " updateexposure=yes filterexposure=yes" 

    write(errstr,*)trim(command)
    call message(AppMsg,Noisy)
    call system (trim(command))
!
! next, make the FOV annulus spectrum (5)
!
!
! already applied pattern, flag and cornerdef, so just energy,
! binsize, etc needed:
!
    fovannspecF = trim(root) // "-annspec.FIT" ! mos_instnum_fovann.pi
    write(errstr,*) "RATIO_METHOD: FOV annulus spectral file =",&
        trim(fovannspecF)
    call message(AppMsg,Noisy)
    write(errstr,*) "RATIO_METHOD: Get Flare discriminator information"
    call message(AppMsg,Noisy)
    write(errstr,*) "RATIO_METHOD: Evselect extract FOV annulus spectrum"
    call message(AppMsg,Noisy)

    command = "evselect table=" // trim(fovannevliF) // &
         " withfilteredset=yes expression='TRUE'"//&
!         "'(PATTERN<=12)&&((FLAG & 0x766a0f63) == 0)&&"//&
!         trim(cornerdef) //  
         " filtertype=expression "//&
         "keepfilteroutput=no updateexposure=yes filterexposure=yes "//&
         "withspectrumset=yes spectrumset=" // trim(fovannspecF) //&
         " energycolumn=PI spectralbinsize=15 withspecranges=yes "//&
         "specchannelmin=0 specchannelmax=11999"
    write(errstr,*)trim(command)
    call message(AppMsg,Noisy)
    call system (trim(command))

! $comm="evselect table=mos".$ins_num."-clean.fits:EVENTS ".
! I don't get this 'clean' file. sounds like an espfilt file
! cleaned but doesn't evselect do the right thing below on the
! raw event file?
!
!!!bhp - use *raw* emchain output eventlist, fovanndef cleans it...
!
!kk get the flare discriminator information
!kk extract the FOV annulus spectrum
!
! original:
! command="backscale spectrumset=mos".$ins_num."-fovann.pi "//&
!         "badpixlocation=mos".$ins_num."-clean.fits withbadpixcorr=yes "//&
!         "ignoreoutoffov=no";
!
! BACKSCAL = area of region used to make spectral file 
! now run backscale on fov annulus spectrum (6)
!
    write(errstr,*) "RATIO_METHOD: Running Backscale on FOV Annulus spectrum."
    call message(AppMsg,Noisy)
!
! fovspec = P0099840101M1S001-annspec.FIT 
!
    command="backscale spectrumset=" // trim(fovannspecF) // &
         " badpixlocation=" // trim(fovannevliF) // &
         " withbadpixcorr=yes ignoreoutoffov=no"
    write(errstr,*)trim(command)
    call message(AppMsg,Noisy)
    call system (trim(command))

! kk get the 8-12keV flux and backscale-the new way
!  bp get it via the OAL
! kk command = "fkeypar " // trim(corspecF) 
! kk get some header keywords to create output file's names (root)
!  bp clean/check here (look for hasblock, etc) 

    set = dataSet(corspecF, READ)
    tab = table(set,"SPECTRUM")
    cor_area = real32Attribute(tab,"BACKSCAL") !corn_scl
    call release(set)
    write(errstr,*) "RATIO_METHOD: Corner Area:",cor_area
    call message(AppMsg,Sparse)
!    print*,'cor_area    = ',cor_area

    set = dataSet(fovannspecF, READ)
    tab = table(set,"SPECTRUM")
    fovann_area = real32Attribute(tab,"BACKSCAL") !annu_scl
    call release(set)
    write(errstr,*) "RATIO_METHOD: FOV annulus Area:",fovann_area
    call message(AppMsg,Sparse)
!    print*,'fovann_area = ',fovann_area
!
    set = dataSet(corevliF, READ)
    tab = table(set,"EVENTS",READ)
    cor_cnt = real(numberOfRows(tab))
    call release(set)
    write(errstr,*) "RATIO_METHOD: Corner total counts:",cor_cnt
    call message(AppMsg,Sparse)
!    print*,'cor_cnt     = ',cor_cnt

    set = dataSet(fovannevliF, READ)
    tab = table(set,"EVENTS",READ)
    fovann_cnt = real(numberOfRows(tab))
    call release(set)
    write(errstr,*) "RATIO_METHOD: FOV annulus total counts:",fovann_cnt
    call message(AppMsg,Sparse)
!    print*,'fovann_cnt  = ',fovann_cnt
!
    set = dataSet(corspecF, READ)
    tab = table(set,"SPECTRUM")
    cor_time = real32Attribute(tab,"EXPOSURE") !time_cor
    call release(set)
    write(errstr,*) "RATIO_METHOD: Corner total exposure:",cor_time
    call message(AppMsg,Sparse)
!    print*,'cor_time    = ',cor_time

    set = dataSet(fovannspecF, READ)
    tab = table(set,"SPECTRUM")
    fovann_time = real32Attribute(tab,"EXPOSURE") !time_ann
    call release(set)
    write(errstr,*) "RATIO_METHOD: Corner total exposure:",fovann_time
    call message(AppMsg,Sparse)
!    print*,'fovann_time = ',fovann_time


! After you create the event files for each of the regions,
! 1. extract a spectrum, and "backscale" it (run backscale). 
! 2. from the header of the spectrum:
!    time = value in exposure keyword
!    area = value in backscal keyword
! Watch out for the area as this is a large number (essentially the number 
! of pixels) so you will need long or something.

!
! What I have done is extract the number of counts per pixel in 8.0-12.0 keV
! in an annulus in the FOV (13900<R<16900) and compare to the counts per pixel
! in the corner regions for the same energy band.
! The ratio of interest is $annu_cnt * $corn_scl/($corn_cnt * $ annu_scl)
! If this ratio is greater than 1.2 then you are definitely flared.
!
    write(errstr,*) "RATIO_METHOD: Calculating ratio:"
    call message(AppMsg,Sparse)
    write(errstr,*) "RATIO_METHOD: ratio = (fovann_cnts/corner_counts) * "
    call message(AppMsg,Sparse)
    write(errstr,*) "RATIO_METHOD: (corner_time*corner_area/corner_counts) / "
    call message(AppMsg,Sparse)
    write(errstr,*) "RATIO_METHOD: (fovann_time*fovann_area)"
    call message(AppMsg,Sparse)
    newratio = fovann_cnt/cor_cnt * &
         (cor_time*cor_area)/(fovann_time*fovann_area)
    unc = sqrt(((fovann_cnt/(cor_cnt**2)+(1/cor_cnt**3))) * &
         ((cor_time*cor_area)/(fovann_time*fovann_area)))
!    newratio = fovann_cnt * cor_area / (cor_cnt * fovann_area)
!    write(errstr,*) "RATIO_METHOD: Input Ratio = ",ratio
!    call message(AppMsg,Noisy)
!    write(errstr,*) "RATIO_METHOD: Calculated Ratio = ",newratio
!    call message(AppMsg,Noisy)
!    print *,' newratio  = ',newratio
!
!    if cnts_ann: number of counts in annulus region
!       area_ann: area of annulus
!       time_ann: exposure time in annulus   
!       cnts_cor: number of counts in corner   
!       area_cor: area of corner 
!       time_cor: exposure time in annulus
!    then 
!       ratio = cnts_ann/cnts_cor * (time_cor*area_cor)/(time_ann*area_ann)
!       unc = sqrt([(cnts_ann/cnts_cor2)+(1/cnts_cor3)])*(time_cor*area_cor)/(time_ann*area_ann)
!
    write(errstr,*) "RATIO_METHOD:                ratio = ",ratio
    call message(AppMsg,Sparse)
    write(errstr,*) "RATIO_METHOD: uncertainty in ratio = ",unc
    call message(AppMsg,Sparse)
!
! Now, make a raw object image
!

    write(errstr,*) "RATIO_METHOD: Running evselect to create "//&
         "*OBJECT IMAGE*"
    call message(AppMsg,Verbose)
    objimgF = trim(root)//"-objimg.FIT"
    command = "evselect table="// trim(evfile) //" withimageset=yes "//&
         " imageset="// trim(objimgF) // " xcolumn=DETX ycolumn=DETY "//&
         "imagebinning=imageSize ximagesize=600 yimagesize=600"
    call system(trim(command))
! 
!!! hmmm....if user selected min and max chan, is that ok? Or will
!that screw up the 8-12k range later???
!
! now make an object LC in the selected range
!
    write(errstr,*) "RATIO_METHOD: Running evselect to create "//&
         "*OBJECT LIGHTCURVE*"
    call message(AppMsg,Verbose)
    objlcF = trim(root)//"-objlc.FIT"
    command = "evselect table="// trim(evfile) //":EVENTS "//&
         "withfilteredset=yes expression='(PATTERN<=" // trim(patt) //&
         ")&&(PI in "//&
         "["// trim(cminchan) // ":" // trim(cmaxchan) // "])&&"//&
         "((FLAG & 0xfb0000) == 0)' filtertype=expression "//&
         "keepfilteroutput=no withrateset=yes rateset="// trim(objlcF) //&
         " timecolumn='TIME' maketimecolumn=yes makeratecolumn=yes "//&
         "timebinsize=1 withtimeranges=no"
    call system(trim(command))
!
! now the corner LC in the selected range
!
    write(errstr,*) "RATIO_METHOD: Running evselect to create "//&
         "*CORNER LIGHTCURVE*"
    call message(AppMsg,Verbose)
    corlcF = trim(root)//"-corlc.FIT"
    command = "evselect table="// trim(evfile) //":EVENTS "//&
         "withfilteredset=yes expression='(PATTERN<=" // trim(patt) //&
         ")&&(PI in "//&
         "["// trim(cminchan) // ":" // trim(cmaxchan) // "])&&"//&
         "((FLAG & 0x766a0f63) == 0)&&"// trim(cornerdef) //&
         "' filtertype=expression keepfilteroutput=no "//&
         "withrateset=yes rateset="// trim(corlcF) //&
         " timecolumn='TIME' maketimecolumn=yes makeratecolumn=yes "//&
         "timebinsize=1 withtimeranges=no"
    call system(trim(command))

!
! now the fov annulus LC in the selected range
!
    write(errstr,*) "RATIO_METHOD: Running evselect to create "//&
         "*FOV annulus LIGHTCURVE*"
    call message(AppMsg,Verbose)
    fovannlcF = trim(root)//"-annlc.FIT"
    command = "evselect table="// trim(evfile) //":EVENTS "//&
         "withfilteredset=yes expression='(PATTERN<=" // trim(patt) //&
         ")&&(PI in "//&
         "["// trim(cminchan) // ":" // trim(cmaxchan) // "])&&"//&
         "((FLAG & 0x766a0f63) == 0)&&"// trim(fovanndef) //&
         "' filtertype=expression keepfilteroutput=no "//&
         "withrateset=yes rateset="// trim(fovannlcF) //&
         " timecolumn='TIME' maketimecolumn=yes makeratecolumn=yes "//&
         "timebinsize=1 withtimeranges=no"
    call system(trim(command))
!
! call PGPLOT directly in the future? Eliminate QDP file construct!!!
!
    histqdpF = trim(root)//"-hist.qdp" ! QDP output file
    write(errstr,*) "RATIO_METHOD: Running Clean_LC..."
    call message(AppMsg,Verbose)
    gtitxtF = trim(root)//"-gti.txt" ! GTI temp text file
    gtifitsF = trim(root)//"-gti.FIT"! GTIBUILD output FITS file
!
! clean_lc is the engine that creates the gti txt file for use by
! gtibuild and the smoothed, binned lc's
!
    method = 2 ! corner = 1, ratio = 2 !!! fix this

    call clean_lc(gtitxtF,objlcF,corlcF,fovannlcF,histqdpF,&
         binning,smoothing,binsize,smoothfactor,method,newratio)

    command = "gtibuild file="// trim(gtitxtF) //" table="// trim(gtifitsF)
    call system(trim(command))
!
! if GTI file exists, create filtered, flare-free event list
!
    inquire(file=trim(gtifitsF),exist=f_ok)
    if (f_ok) then
       objevlifiltF = trim(root)//"-objevlifilt.FIT"
       command = "evselect table="// trim(evfile) //&
            " withfilteredset=yes "//&
            "expression='(PATTERN<=" // trim(patt) //")&&GTI("//&
            trim(gtifitsF) //",TIME)&&"//&
            "((FLAG & 0x766a0f63)==0)' filteredset="//&
            trim(objevlifiltF) //&
            " filtertype=expression keepfilteroutput=yes "//&
            "updateexposure=yes filterexposure=yes"
       call system(trim(command))
       objimgfiltF = trim(root)//"-objimgfilt.FIT"
       command = "evselect table="// trim(objevlifiltF) //&
            " withimageset=yes imageset="// trim(objimgfiltF) //&
            " xcolumn=DETX ycolumn=DETY imagebinning=imageSize "//&
            "ximagesize=600 yimagesize=600"
       call system(trim(command))
!
! create corner-only events file 
!
       corevlifiltF = trim(root) // "-corevlifilt.FIT"
       command = "evselect table="// trim(evfile) //":EVENTS " //&
            "withfilteredset=yes expression='(PATTERN<=" // trim(patt) //&
            ")&&GTI("// trim(gtifitsF) //&
            ",TIME)&&((FLAG & 0x766a0f63) == 0)&&"//&
            trim(cornerdef)//"' filteredset="// trim(corevlifiltF)//&
            " filtertype=expression keepfilteroutput=yes "//&
            "updateexposure=yes filterexposure=yes" 
       call system(trim(command))

! create the image of the corner-only events for mos_doc.bat use
       corimgfiltF = trim(root) // "-corimgfilt.FIT"
       command = "evselect table="// trim(corevlifiltF) //&
            " withimageset=yes imageset="// trim(corimgfiltF) //&
            " xcolumn=DETX ycolumn=DETY imagebinning=imageSize "//&
            "ximagesize=600 yimagesize=600"
       call system(trim(command))
       
    else
       call error("noGTI","No GTI FITS file was found, cannot continue")
    end if

    return

  end subroutine ratio_method
!
!
!
  subroutine clean_lc(gtitxtFile,objlcFile,corlcFile,fovlcFile,&
       histqdpFile,binning,smoothing,binsize,smoothfactor,&
       method,ratio)
    type(TableT) :: objlcTbl, corlcTbl, fovlcTbl
    type(ColumnT) :: time_lcCol, rate_lcCol
    type(ColumnT) :: time_clcCol, rate_clcCol
    type(ColumnT) :: time_flcCol, rate_flcCol
    type(DataSetT) :: objlcS, corlcS, fovlcS
    character(len=1024) :: objlcFile, corlcFile, fovlcFile
    character(len=1024) :: histqdpFile, gtitxtFile
    character(len=10) :: obsid
    logical, intent(in) :: binning, smoothing
    integer :: binwidth, smoo, ierr
    integer, intent(in) :: binsize, smoothfactor, method
    integer :: nbins_lc, nbins_clc, nchan_lc, nchan_clc,&
         nbins_flc, nchan_flc
    integer :: low = 0, high = 0
    integer :: i, ii, n, nn, np
    real(kind=double), intent(in) :: ratio
    real(kind=double) :: maxr, maxp, th, rlow, rhigh
    real(kind=double) :: delt = 12.0
    real(kind=double) :: hist_cnts(10000), hist(10000), unc(10000),&
         chan(10000), fit(10000), ap(10)
    real(kind=double), allocatable, dimension(:) :: &
         tstart_lc, tend_lc, tstart_clc, tend_clc, &
         tstart_flc, tend_flc, new_modrate_lc, new_modrate_clc, &
         new_modrate_flc, modrate_lc, modrate_clc, modrate_flc, &
         time_lc, time_clc, time_flc, rate_lc, rate_clc, rate_flc
    real(kind=double), dimension(:), pointer   :: &
         p_time_lc, p_rate_lc, p_time_clc, p_rate_clc,&
         p_time_flc, p_rate_flc
    ap = (/0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0/)
    rlow = 0.
    rhigh = 0.
!
! obsid is in the file name
!
    obsid = objlcFile(2:11)
!   
! open and read the object lightcurve
!
    objlcS = dataSet(trim(objlcFile), READ)
    if (.not.hasBlock(objlcS,"RATE")) then
       call error("badInfile","Object LC does not contain a RATE extension")
    else
       objlcTbl = table(objlcS,"RATE")
    endif
    if (.not.hasColumn(objlcTbl,"TIME")) then
       call error("badTable","RATE table does not contain a TIME column")
    else
       time_lcCol = column(objlcTbl,"TIME",READ)
       p_time_lc => Real64Data(time_lcCol)
    endif
    if (.not.hasColumn(objlcTbl,"RATE")) then
       call error("badTable","RATE table does not contain a RATE column")
    else
       rate_lcCol = column(objlcTbl,"RATE",READ)
       p_rate_lc => Real64Data(rate_lcCol)
    endif
    nchan_lc = numberOfrows(objlcTbl)
    write(errstr,*) "CLEAN_LC: There are ",nchan_lc," rows in the Object LC"
    call message(AppMsg,Verbose)
!    print *,' There are ',nchan_lc, 'rows in objlc'
!
! open and read the corner lightcurve (!!! use data subspace???
!
    corlcS = dataSet(trim(corlcFile), READ)
    if (.not.hasBlock(corlcS,"RATE")) then
       call error("badInfile","Corner LC does not contain an RATE extension")
    else
       corlcTbl = table(corlcS,"RATE")
    endif
    if (.not.hasColumn(corlcTbl,"TIME")) then
       call error("badTable","RATE table does not contain a TIME column")
    else
       time_clcCol = column(corlcTbl,"TIME",READ)
       p_time_clc => Real64Data(time_clcCol)
    endif
    if (.not.hasColumn(corlcTbl,"RATE")) then
       call error("badTable","RATE table does not contain a RATE column")
    else
       rate_clcCol = column(corlcTbl,"RATE",READ)
       p_rate_clc => Real64Data(rate_clcCol)
    endif
    nchan_clc = numberOfrows(corlcTbl)
    write(errstr,*) "CLEAN_LC: There are ",nchan_clc," rows in the Corner LC."
    call message(AppMsg,Verbose)
!    print *,' There are ',nchan_clc, 'rows in corlc'
!
! set up binning and smoothing 
!
    if (.not.binning) then
       binwidth = 1
       write(errstr,*) "CLEAN_LC: Binning factor of 1."
       call message(AppMsg,Verbose)    
    else 
       binwidth = binsize
    end if
    if (.not.smoothing) then
       smoo = 1
       write(errstr,*) "CLEAN_LC: Smoothing factor of 1."
       call message(AppMsg,Verbose)    
    else 
       smoo = smoothfactor
    end if
!
! do the below only for RATIO METHOD
!
!    print *,' got to here'
!    print *,' method is ', method
!    print *,' fovlcFile is', trim(fovlcFile)
    if (method .eq. 2) then 
       fovlcS = dataSet(trim(fovlcFile), READ)
       if (.not.hasBlock(fovlcS,"RATE")) then
          call error("badInfile",&
               "FOV annulus LC does not contain an RATE extension")
       else
          fovlcTbl = table(fovlcS,"RATE")
       endif
       if (.not.hasColumn(fovlcTbl,"TIME")) then
          call error("badTable","RATE table does not contain a TIME column")
       else
          time_flcCol = column(fovlcTbl,"TIME",READ)
          p_time_flc => Real64Data(time_flcCol)
       endif
       if (.not.hasColumn(fovlcTbl,"RATE")) then
          call error("badTable","RATE table does not contain a RATE column")
       else
          rate_flcCol = column(fovlcTbl,"RATE",READ)
          p_rate_flc => Real64Data(rate_flcCol)
       endif
       nchan_flc = numberOfrows(fovlcTbl)
       write(errstr,*) "CLEAN_LC: There are ",nchan_flc,&
            " rows in the FOV annulus LC."
       call message(AppMsg,Verbose)
!       print *,' There are ',nchan_flc, 'rows in fovlc'
!
       allocate(time_flc(nchan_flc))
       allocate(rate_flc(nchan_flc))
       time_flc = p_time_flc
       rate_flc = p_rate_flc
       allocate(modrate_flc(nchan_flc))
       allocate(tstart_flc(nchan_flc))
       allocate(tend_flc(nchan_flc))
       write(errstr,*) &
            "CLEAN_LC: Calling binit to bin up the FOV annulus LC."
       call message(AppMsg,Verbose)
       call binit(time_flc, rate_flc, binwidth, nchan_flc, &
            modrate_flc,tstart_flc, tend_flc, nbins_flc)
       nbins_flc = nbins_flc - 1        
       do while (tend_flc(nbins_flc) .le. tstart_flc(nbins_flc))
          nbins_flc = nbins_flc - 1
       enddo
       write(errstr,*) "CLEAN_LC: Smoothing the FOV annulus LC"
       call message(AppMsg,Noisy)
       allocate(new_modrate_flc(nchan_flc))
       call smooth(rate_flc,nchan_flc,smoo,new_modrate_flc)
    else
!
! dummy pointers for corner method. FIX THIS!!!
!
       allocate(time_flc(1))
       allocate(rate_flc(1))
       allocate(modrate_flc(1))
       allocate(tstart_flc(nchan_flc))
       allocate(tend_flc(nchan_flc))
    end if
    write(errstr,*) "CLEAN_LC: Calling binit to bin up the Object LC."
    call message(AppMsg,Verbose)
    allocate(time_lc(nchan_lc))
    allocate(rate_lc(nchan_lc))
    time_lc = p_time_lc
    rate_lc = p_rate_lc
    allocate(modrate_lc(nchan_lc))
    allocate(tstart_lc(nchan_lc))
    allocate(tend_lc(nchan_lc))
!
    write(errstr,*) "CLEAN_LC: Calling binit to bin up the Corner LC."
    call message(AppMsg,Verbose)
    allocate(time_clc(nchan_clc))
    allocate(rate_clc(nchan_clc))
    time_clc = p_time_clc
    rate_clc = p_rate_clc
    allocate(modrate_clc(nchan_clc))
    allocate(tstart_clc(nchan_clc))
    allocate(tend_clc(nchan_clc))
!
! modrate is used for plotting, new_modrate is the smoothed
!
    write(errstr,*) "CLEAN_LC: Calling binit to bin up the Object LC."
    call message(AppMsg,Verbose)
    call binit(time_lc, rate_lc, binwidth, nchan_lc, modrate_lc, &
         tstart_lc, tend_lc, nbins_lc)
    nchan_lc = nchan_lc - 1       

    write(errstr,*) "CLEAN_LC: Calling binit to bin up the Corner LC."
    call message(AppMsg,Verbose)
    call binit(time_clc, rate_clc, binwidth, nchan_clc, modrate_clc, &
         tstart_clc, tend_clc, nbins_clc)

    nbins_lc = nbins_lc - 1 
    nbins_clc = nbins_clc - 1 

    do while (tend_lc(nbins_lc) .le. tstart_lc(nbins_lc))
       nbins_lc = nbins_lc - 1
    enddo
    do while (tend_clc(nbins_clc) .le. tstart_clc(nbins_clc))
       nbins_clc = nbins_clc - 1
    enddo
!
    n = nbins_lc ! doesn't get used?!
!
!  Smooth the data for fitting the distribution
!
    write(errstr,*) "CLEAN_LC: Smoothing the Object LC"
    call message(AppMsg,Noisy)
    allocate(new_modrate_lc(nchan_lc))
    call smooth(rate_lc,nchan_lc,smoo,new_modrate_lc)
    write(errstr,*) "CLEAN_LC: Smoothing the Corner LC"
    call message(AppMsg,Noisy)
    allocate(new_modrate_clc(nchan_clc))
    call smooth(rate_clc,nchan_clc,smoo,new_modrate_clc)
!
!  CORNER_METHOD: Make the histogram, use the smoothed data
!
    do i=1,10000
       hist_cnts(i) = 0.0
    enddo
       
    do i=1,nchan_lc
       nn = 50.0 * new_modrate_lc(i) + 1
       if(nn .le. 10000) hist_cnts(nn) = hist_cnts(nn) + 1.0
    enddo
!
!  find the peak of the distribution
!
    write(errstr,*) "CLEAN_LC: Finding the peak of the distribution."
    call message(AppMsg,Noisy)
    
    maxr = 0.0
    maxp = 0
    do i=10,9990
       th = 0.0
       do ii =i-2,i+2
          th = th + hist_cnts(ii)
       enddo
       th = th/5.0
       fit(i) = th
       if (th .ge. maxr) then
          maxr = th
          maxp = i
       endif
    enddo
    do i=10,9990
       hist_cnts(i) = fit(i)
    enddo
!
!  Copy the data over to the fit arrays
!
    delt = sqrt(maxp)*delt/5.0
    low = maxp - delt
    if(low .lt. 1) low = 1
    high = int(maxp + delt)
    if(high .gt. 10000) high = 10000
    nn = 0
    do i=low,high
       nn = nn + 1
       chan(nn) = i
       hist(nn) = hist_cnts(i)
       unc(nn) = sqrt(hist_cnts(i))
       if(unc(nn) .lt. 1) unc(nn) = 1
    enddo
    rlow = low/50.0
    rhigh = high/50.0
    np = 3
    ap(1) = maxr
    ap(2) = 1.0/delt
    ap(3) = maxp
    write(errstr,*) "CLEAN_LC: maxp = ",maxp," low = ",low," high = ",high
    call message(AppMsg,Verbose)
    
    write(errstr,*) "CLEAN_LC: Initial Guess:", (ap(i),i=1,3), nn, np
    call message(AppMsg,Verbose)
!
! still do gaussian fit even if using ratio method. Need values
! for ap for plots! RATIO_METHOD subroutine filters data..
!
!    if (method .eq. 1) then 
       call gauss_fit(chan,hist,unc,nn,np,ap,ierr)
!    end if
    write(errstr,*) "CLEAN_LC: Final Values:", (ap(i),i=1,3), nn, np
    call message(AppMsg,Verbose)
!    
!  write out a QDP plot file and fill GTI for filtering
!
    write(errstr,*) "CLEAN_LC: Writing out plots to QDP file: ",&
         trim(histqdpFile)
    call message(AppMsg,Verbose)
!
    call write_qdp_gti(obsid,histqdpFile,gtitxtFile,ap,hist_cnts,&
         tstart_lc,tend_lc,tstart_clc,tend_clc,tstart_flc,tend_flc,&
         nbins_lc,nbins_clc,nbins_flc,modrate_lc,modrate_clc,&
         modrate_flc,rlow,rhigh,ierr,method,ratio)
!    deallocate (hist_cnts)
    deallocate (tstart_lc)
    deallocate (tend_lc)
    deallocate (tstart_clc)
    deallocate (tend_clc)
    deallocate (tstart_flc)
    deallocate (tend_flc)
    deallocate (modrate_lc)
    deallocate (modrate_clc)
    deallocate (modrate_flc)
    return
  end subroutine clean_lc
!
!
!
  subroutine binit (time,rate,binsize,nchan,bin_rate,tstart,tend,n)
!
! simple binning routine
!
    integer :: i, nn
    integer, intent(in) :: binsize, nchan
    integer, intent(out) :: n
    real(kind=double), intent(in) :: time(nchan), rate(nchan)
    real(kind=double), intent(out) :: &
         bin_rate(nchan), tstart(nchan), tend(nchan)
    real(kind=double) :: t, c

    t = time(1)
    c = rate(1)
    i = 2
    n = 1
    tstart(n) = time(1) - 0.5
    nn = 1

    do while (i .lt. nchan)
       do while ((time(i) - t .lt. binsize) .and. (i .le. nchan-1))
          c = c + rate(i)
          nn = nn + 1
          i = i + 1
       enddo
       if(nn .gt. 0) then
          bin_rate(n) = c/nn
          tend(n) = time(i-1) + 0.5
          n = n + 1
       endif
       tstart(n) = time(i) - 0.5
       c = 0.0
       nn = 0
       t = time(i)
    end do
    return
  end subroutine binit
!
!
!
  subroutine smooth(rate,nchan,smoo,smoo_rate)
!
! simple smoothing routine
!
    implicit none

    integer, intent(in) :: nchan, smoo
    integer :: i, ih, j, j1, j2
    real(kind=double) :: n, t = 0.
    real(kind=double), intent(in) :: rate(nchan)
    real(kind=double), intent(out) :: smoo_rate(nchan)

    ih = smoo/2
    do i=1,nchan
       j1 = i - ih
       if(j1 .lt. 1) j1 = 1
       j2 = i + ih
       if(j2 .gt. nchan) j2 = nchan
       t = 0.0
       n = 0.0
       do j=j1,j2
          t = t + rate(j)
          n = n + 1.0
       enddo
       smoo_rate(i) = t/n
    enddo
    return
  end subroutine smooth
!
!  
  subroutine write_qdp_gti(obsid,histqdpF,gtitxtF,ap,hist_cnts,&
       tstart_lc,tend_lc,tstart_clc,tend_clc,tstart_flc,tend_flc,&
       nbins_lc,nbins_clc,nbins_flc,modrate_lc,modrate_clc,&
       modrate_flc,rlow,rhigh,ierr,method,ratio)
!
! write histo, rate, information to QDP file for later plotting
! should be changed to plot directly to PS! 
!
    character(len=10), intent(in) :: obsid
    character(len=1024) :: histqdpF, gtitxtF
    integer, intent(in) :: nbins_lc, nbins_clc, nbins_flc, method
    integer :: i, ii, ierr, n, nn, status, clobber, lun, pflg = 1
    real(kind=double), intent(in), dimension(10) :: ap
    real(kind=double), intent(in), dimension(:) :: hist_cnts
    real(kind=double), intent(in), dimension(:) :: tstart_lc, tend_lc, &
         tstart_clc, tend_clc, tstart_flc, tend_flc, &
         modrate_lc, modrate_clc, modrate_flc
    real(kind=double) :: ratelimm, ratelimp, x, dt, y, ys
    real(kind=double) :: xmax
    real(kind=double) :: d1 = -1.0, d2 = 0.0, sca = 1.5, ymax, ymaxf
    real(kind=double), allocatable, dimension(:) :: tstart, tend
    real(kind=double) :: rlow, rhigh, rat
    real(kind=double), intent(in) :: ratio
    clobber = 1
    status = 0
    call ftgiou(lun,status)
    if(status .ne. 0) then
       call warning("NoOpenQDPfile","WRITE_QDP_GTI: Unable to open QDP file for output.")
    endif
!    call warning ("obsid = ",trim(obsid))
    if(clobber .eq. 1) then
       open(unit=lun,status='unknown',form='formatted',&
            file=trim(histqdpF))
    else
       open(unit=lun,status='new',form='formatted',&
            file=trim(histqdpF))
    endif
    write(lun,*) 'viewport 0.05 0.05 0.98 0.89'
    write(lun,*) 'font roman'
    write(lun,*) 'cs 1.0'
    write(lun,*) 'skip sing'
    write(lun,*) 'gap errors'
    Write(lun,*) 'col 1 on 1'
    write(lun,*) 'col 3 on 2'
    write(lun,*) 'col 1 on 3'
    write(lun,*) 'col 3 on 4'
    write(lun,*) 'col 1 on 5'
    write(lun,*) 'col 3 on 6'
    write(lun,*) 'col 2 on 7'
    write(lun,*) 'col 4 on 8'
    write(lun,*) 'line step 2'
    write(lun,*) 'read serr 2'
    write(lun,*) 'read serr 4'
    write(lun,*) 'read serr 5'
    write(lun,*) 'read serr 6'
    write(lun,*) 'read serr 7'
    write(lun,*) 'win 1'
    write(lun,*) 'yplot 1 2'
    write(lun,*) 'loc 0.03 0.72 0.97 1.0'
    write(lun,*) 'lab x Count Rate (counts/s)'
!    write(lun,*) 'lab f'
    write(lun,*) 'la y N'
    xmax = 2.0*ap(3)/50.0
    if(xmax .lt. 5.0) xmax = 5.0
    write(lun,*) 'r x1 0 ', xmax
    ymax = 0.0
    do i=1,10000
       y = ap(1)*exp(-ap(2)*(i - ap(3))*(i - ap(3)))
       if(hist_cnts(i) + 1.1*sqrt(hist_cnts(i)) .gt. ymax) &
            ymax = hist_cnts(i) + 1.1*sqrt(hist_cnts(i))
    enddo
    ymax = 1.1*ymax
    ymaxf = ymax
    write(errstr,*) &
         "WRITE_QDP_GTI: Ymax = ",ymax
    call message(AppMsg,Verbose)
    write(lun,*) 'r y1 0 ', ymax
    write(lun,*) 'la 1 p ', 0.5*xmax, 0.9*ymax,&
         ' "Count Rate Histogram"'
!    if(ap(3) .lt. 120.0) then
!       write(lun,*) 'la 11 p  4.8 ', 0.9*ymax,&
!            ' cs 0.8 just right "Fit Limits: Blue"'
!       write(lun,*) 'la 12 p  4.8 ', 0.8*ymax,&
!            ' cs 0.8 just right "Selection Limits: Red"'
!       write(lun,*) 'la 13 p ', 0.8*xmax, 1.05*ymax,&
!            ' cs 0.8 just right "OBSID: ', trim(obsid), '"'
!    else
       write(lun,*) 'la 11 p ', 0.04*xmax, 0.9*ymax,&
            ' cs 0.8 just left "Fit Limits: Blue"'
       write(lun,*) 'la 12 p ', 0.04*xmax, 0.8*ymax,&
            ' cs 0.8 just left "Selection Limits: Red"'
       write(lun,*) 'la 13 p ', 0.8*xmax, 1.05*ymax,&
            ' cs 0.8 just left "ObsID: ',obsid, '"'
       write(lun,3008) 0.78*xmax, 0.9*ymax, ap(1)
3008   format('la 14 p ',f10.4,f10.2,' cs 0.8 just ',&
            'left "Fit Norm: ',f7.1,'"')
       write(lun,3009) 0.78*xmax, 0.8*ymax, 0.02*sqrt(1.0/ap(2))
3009   format('la 15 p ',f10.4,f10.2,' cs 0.8 just ',&
           'left "Fit Width: ',f7.3,' c/s"')
       write(lun,3010) 0.78*xmax, 0.7*ymax, (0.02*ap(3)-0.01)
3010   format('la 16 p ',f10.4,f10.2,' cs 0.8 just ',&
           'left "Fit Center: ',f7.3,' c/s"')
!    endif
!
!  Write out the control lines for the plot of the FOV light curve
!
    write(errstr,*) &
         "WRITE_QDP_GTI: Write out the control lines for the plot of the FOV light curve"
    call message(AppMsg,Noisy)
    write(lun,*) 'win 2'
    write(lun,*) 'yplot 3 4'
    write(lun,*) 'loc 0.03 0.39 0.97 0.67'
    write(lun,*) 'la x Time (s)'
    write(lun,*) 'la y Count Rate (counts/s)'
    write(lun,*) 'r x2 0 ', tend_lc(nbins_lc) - tstart_lc(1)
    ymax = 0.0
    do i=1,nbins_lc
       dt = tend_lc(i) - tstart_lc(i)
1011   format(1x,a3,f8.3,a6,i5,a2,f12.1,a8,f12.1,a9,f12.1)
       if(modrate_lc(i) + sqrt(modrate_lc(i)*dt)/dt .gt. ymax) then
          ymax = modrate_lc(i) + sqrt(modrate_lc(i)*dt)/dt
       endif
    enddo
    write(lun,*) 'r y2 0.1 ', ymax
    write(lun,*) 'la 2 p ', 0.5*(tend_lc(nbins_lc)-tstart_lc(1)),& 
         0.5*ymax,' "FOV Light Curve"'
    write(lun,*) 'log y'

!
!  Write out the control lines for the plot of the corner light curve
!
    write(errstr,*) &
         "WRITE_QDP_GTI: Write out the control lines for the plot of the Corner light curve"
    call message(AppMsg,Noisy)
    write(lun,*) 'win 3'
    write(lun,*) 'yplot 5 6'
    write(lun,*) 'loc 0.03 0.06 0.97 0.34'
    write(lun,*) 'la x Time (s)'
    write(lun,*) 'la y Count Rate (counts/s)'
    write(lun,*) 'r x3 0 ', tend_clc(nbins_clc)-tstart_clc(1)
    ymax = 0.0
    do i=1,nbins_clc
       dt = tend_clc(i) - tstart_clc(i)
       if(modrate_clc(i) + sqrt(modrate_clc(i)*dt)/dt .gt. ymax) then
          ymax = modrate_clc(i) + sqrt(modrate_clc(i)*dt)/dt
       endif
    enddo
    write(lun,*) 'r y3 0.0 ', ymax
    write(lun,*) 'la 3 p ', 0.5*(tend_clc(nbins_clc)-tstart_clc(1)),&
         0.9*ymax,' "Corner Light Curve"'
!
    do i=1,10000
       x = 0.02*i - 0.01
       y = ap(1)*exp(-ap(2)*(i - ap(3))*(i - ap(3)))
       write(lun,2000) x, hist_cnts(i), sqrt(hist_cnts(i)),&
            y, d1, d2, d1, d2, d1, d2, d1, d2, d1, d1
2000   format(1x,f12.3,3f10.3,10f5.1)
    enddo
!
!  Set up the good time intervals
!
    write(errstr,*) "WRITE_QDP_GTI: Set up GTIs"
    call message(AppMsg,Noisy)
    ratelimm = 0.02*(ap(3) - sca/sqrt(ap(2)))
    ratelimp = 0.02*(ap(3) + sca/sqrt(ap(2)))
    nn = 0
    allocate(tstart(nbins_lc))
    allocate(tend(nbins_lc))
    do i=1,nbins_lc
       if((modrate_lc(i) .gt. 0.0) .and. (modrate_lc(i) .le. ratelimp) &
            .and. (modrate_lc(i) .ge. ratelimm)) then
          nn = nn + 1
          tstart(nn) = tstart_lc(i)
          tend(nn) = tend_lc(i)
       endif
    enddo
!
!  If the intervals are close then run them together
!
    n = 1
    do i=2,nn
       if(abs(tstart(i) - tend(i-1)) .gt. 5.0) then
          tend(n) = tend(i-1)
          n = n + 1
          tstart(n) = tstart(i)
       endif
    enddo
    if (ierr.gt.0) then 
       if (nn.eq.0) nn = 1   ! singular matrix solution
       call warning("SingularMatrix","No Count Rate Histogram Produced")
    end if
   tend(n) = tend(nn)
!
!  Eliminate the short time intervals
!
    nn = 0
    do i=1,n
       if(tend(i)-tstart(i) .gt. 100.0) then
          nn = nn + 1
          tstart(nn) = tstart(i)
          tend(nn) = tend(i)
       endif
    enddo
!
!  Write out a count rate plot file
!
    write(errstr,*) "WRITE_QDP_GTI: Writing out an OBJECT count rate plot file."
    call message(AppMsg,Noisy)

    if(pflg .eq. 1) then
       do i=1,nbins_lc
          x = (tstart_lc(i) + tend_lc(i))/2.0
          dt = tend_lc(i) - tstart_lc(i)
          y = modrate_lc(i)
          ys = sqrt(modrate_lc(i)*dt)/dt
          write(lun,2010) x-tstart_lc(1), d1, d2, d1,& 
               y, ys, d1, d2, d1, d2, d1, d2, d1, d1
2010      format(1x,f12.3,3f5.1,2f10.3,8f5.1)
       enddo
       do i=1,nbins_lc
          x = (tstart_lc(i) + tend_lc(i))/2.0
          do ii=1,nn
             if((x .ge. tstart(ii)) .and. (x .le. tend(ii))) then
                dt = tend_lc(i) - tstart_lc(i)
                y = modrate_lc(i)
                ys = sqrt(modrate_lc(i)*dt)/dt
                write(lun,2020) x-tstart_lc(1), d1, d2, d1,&
                     d1, d2, y, ys, d1, d2, d1, d2, d1, d1
2020            format(1x,f12.3,5f5.1,2f10.3,6f5.1)
             endif
          enddo
       enddo
!
!  Plot the corner data
!
       write(errstr,*) "WRITE_QDP_GTI: Writing out a CORNER count rate plot file."
       call message(AppMsg,Noisy)
       do i=1,nbins_clc
          x = (tstart_clc(i) + tend_clc(i))/2.0
          dt = tend_clc(i) - tstart_clc(i)
          y = modrate_clc(i)
          ys = sqrt(modrate_clc(i)*dt)/dt
          write(lun,2030) x-tstart_clc(1), d1, d2, d1,&
               d1, d2, d1, d2, y, ys, d1, d2, d1, d1 
2030      format(1x,f12.3,7f5.1,2f10.3,4f5.1)
       enddo

       do i=1,nbins_clc
          x = (tstart_clc(i) + tend_clc(i))/2.0
          do ii=1,nn
             if((x .ge. tstart(ii)) .and. (x .le. tend(ii))) then
                dt = tend_clc(i) - tstart_clc(i)
                y = modrate_clc(i)
                ys = sqrt(modrate_clc(i)*dt)/dt
                write(lun,2040) x-tstart_clc(1), d1, d2, d1,&
                     d1, d2, d1, d2, d1, d2, y, ys, d1, d1
2040            format(1x,f12.3,9f5.1,2f10.3,2f5.1)
             endif
          enddo
       enddo
       write(lun,2050) ratelimm, d1, d2, d1,& 
            d1, d2, d1, d2, d1, d2, d1, d2, d2, d1
       write(lun,2050) ratelimm, d1, d2, d1,& 
            d1, d2, d1, d2, d1, d2, d1, d2, 1.05*ymaxf, d1
       write(lun,2050) ratelimp, d1, d2, d1,&
            d1, d2, d1, d2, d1, d2, d1, d2, 1.05*ymaxf, d1
       write(lun,2050) ratelimp, d1, d2, d1,&
            d1, d2, d1, d2, d1, d2, d1, d2, d2, d1
2050   format(1x,f12.3,11f5.1,f10.3,f5.1)
       write(lun,2060) rlow, d1, d2, d1,&
            d1, d2, d1, d2, d1, d2, d1, d2, d1, d2
       write(lun,2060) rlow, d1, d2, d1,& 
            d1, d2, d1, d2, d1, d2, d1, d2, d1, 1.05*ymaxf
       write(lun,2060) rhigh, d1, d2, d1,& 
            d1, d2, d1, d2, d1, d2, d1, d2, d1, 1.05*ymaxf
       write(lun,2060) rhigh, d1, d2, d1,&
            d1, d2, d1, d2, d1, d2, d1, d2, d1, d2
       close(lun)
2060   format(1x,f12.3,12f5.1,f10.3)
    endif
!
!  Write out an ascii file of the gtis
!
    status = 0
    call ftgiou(lun,status)
    if(status .ne. 0) then
       write(errstr,*) "WRITE_QDP_GTI: Unable to write to GTI ASCII file."
       call message(AppMsg,Verbose)
    endif
    if(clobber .eq. 1) then
       open(lun,status='unknown',form='formatted',file=trim(gtitxtF))
    else
       open(lun,status='new',form='formatted',file=trim(gtitxtF))
    endif
    do i=1,nn
       write(lun,*) tstart(i), tend(i), '  +'
    enddo
    close(lun)
    return
  end subroutine write_qdp_gti
  
end module espfilt_mod
